<?php

/**
 * Copyright © 2016 Filoblu S.r.l. All rights reserved.
 */

namespace FiloBlu\Esb\Model\Rest;

use Exception;
use FiloBlu\Esb\Api\Data\AcknowledgeInterface;
use FiloBlu\Esb\Api\ExternalShipmentInterface;
use FiloBlu\Esb\Helper\LoggerProvider;
use Magento\Catalog\Model\Product\Type;
use Magento\Framework\Api\SearchCriteriaBuilder;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\App\ObjectManager;
use Magento\Framework\App\ProductMetadataInterface;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\Exception\LocalizedException;
use Magento\Framework\Exception\MailException;
use Magento\Framework\Phrase;
use Magento\InventoryApi\Api\Data\StockInterface;
use Magento\InventorySales\Model\ResourceModel\StockIdResolver;
use Magento\InventorySalesApi\Api\Data\SalesChannelInterface;
use Magento\InventorySourceSelectionApi\Api\Data\InventoryRequestInterfaceFactory;
use Magento\InventorySourceSelectionApi\Api\Data\ItemRequestInterfaceFactory;
use Magento\InventorySourceSelectionApi\Api\GetDefaultSourceSelectionAlgorithmCodeInterface;
use Magento\InventorySourceSelectionApi\Api\SourceSelectionServiceInterface;
use Magento\Sales\Api\Data\OrderExtensionFactory;
use Magento\Sales\Api\Data\OrderInterface;
use Magento\Sales\Api\Data\ShipmentExtensionInterface;
use Magento\Sales\Api\Data\ShipmentInterface;
use Magento\Sales\Api\Data\ShipmentTrackInterface;
use Magento\Sales\Api\Data\ShipmentTrackInterfaceFactory;
use Magento\Sales\Api\ShipmentRepositoryInterface;
use Magento\Sales\Exception\CouldNotShipException;
use Magento\Sales\Model\Convert\Order;
use Magento\Sales\Model\Order\Config as OrderConfig;
use Magento\Sales\Model\Order\OrderStateResolverInterface;
use Magento\Sales\Model\Order\Shipment\OrderRegistrarInterface;
use Magento\Sales\Model\Order\ShipmentFactory;
use Magento\Sales\Model\OrderRepository;
use Magento\Shipping\Model\ShipmentNotifier;
use Magento\Store\Model\StoreManagerInterface;
use Monolog\Logger;

/**
 * REST Api to save external shipments
 */
class ExternalShipment implements ExternalShipmentInterface
{

    /**
     * @var OrderRepository
     */
    protected $orderRepository;

    /**
     * @var ShipmentFactory
     */
    protected $shipmentFactory;

    /**
     * @var SearchCriteriaBuilder
     */
    protected $searchCriteriaBuilder;

    /**
     * @var ShipmentFactory
     */
    protected $shipmentNotifier;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var OrderExtensionFactory
     */
    protected $orderExtensionFactory;

    /**
     * @var Logger
     */
    protected $logger;

    /**
     * @var AcknowledgeInterface
     */
    protected $response;

    /**
     * @var ShipmentRepositoryInterface
     */
    private $shipmentRepository;

    /**
     * @var ProductMetadataInterface
     */
    private $productMetadata;

    /**
     * @var ShipmentTrackInterfaceFactory
     */
    private $shipmentTrackInterfaceFactory;
    /**
     * @var ResourceConnection
     */
    private $resourceConnection;
    /**
     * @var OrderRegistrarInterface
     */
    private $orderRegistrar;
    /**
     * @var OrderStateResolverInterface
     */
    private $orderStateResolver;
    /**
     * @var OrderConfig
     */
    private $config;

    /**
     * Init plugin
     *
     * @param ScopeConfigInterface $scopeConfig
     * @param OrderRepository $orderRepository
     * @param ShipmentFactory $shipmentFactory
     * @param OrderExtensionFactory $orderExtensionFactory
     * @param ShipmentNotifier $shipmentNotifier
     * @param SearchCriteriaBuilder $searchCriteriaBuilder
     * @param LoggerProvider $loggerProvider
     * @param ShipmentRepositoryInterface $shipmentRepository
     * @param ProductMetadataInterface $productMetadata
     * @param AcknowledgeInterface $response
     * @param ShipmentTrackInterfaceFactory $shipmentTrackInterfaceFactory
     * @param OrderRegistrarInterface $orderRegistrar
     * @param OrderStateResolverInterface $orderStateResolver
     * @param OrderConfig $config
     * @param ResourceConnection $resourceConnection
     */
    public function __construct(
        ScopeConfigInterface $scopeConfig,
        OrderRepository $orderRepository,
        ShipmentFactory $shipmentFactory,
        OrderExtensionFactory $orderExtensionFactory,
        ShipmentNotifier $shipmentNotifier,
        SearchCriteriaBuilder $searchCriteriaBuilder,
        LoggerProvider $loggerProvider,
        ShipmentRepositoryInterface $shipmentRepository,
        ProductMetadataInterface $productMetadata,
        AcknowledgeInterface $response,
        ShipmentTrackInterfaceFactory $shipmentTrackInterfaceFactory,
        OrderRegistrarInterface $orderRegistrar,
        OrderStateResolverInterface $orderStateResolver,
        OrderConfig $config,
        ResourceConnection $resourceConnection
    ) {
        $this->scopeConfig = $scopeConfig;
        $this->orderRepository = $orderRepository;
        $this->shipmentFactory = $shipmentFactory;
        $this->orderExtensionFactory = $orderExtensionFactory;
        $this->shipmentNotifier = $shipmentNotifier;
        $this->searchCriteriaBuilder = $searchCriteriaBuilder;
        $this->logger = $loggerProvider->getLogger();
        $this->response = $response;
        $this->shipmentRepository = $shipmentRepository;
        $this->productMetadata = $productMetadata;
        $this->shipmentTrackInterfaceFactory = $shipmentTrackInterfaceFactory;
        $this->resourceConnection = $resourceConnection;
        $this->orderRegistrar = $orderRegistrar;
        $this->orderStateResolver = $orderStateResolver;
        $this->config = $config;
    }

    /**
     * Save External Shipment data. Save PDF file and additional data on Sales Order.
     *
     * @param ShipmentInterface $shipment_data
     * @return AcknowledgeInterface
     * @throws Exception
     */
    public function save(ShipmentInterface $shipment_data)
    {
        $shipmentExtensionAttributes = $shipment_data->getExtensionAttributes();

        if (null === $shipmentExtensionAttributes || !$shipmentExtensionAttributes->getOrderIncrementId()) {
            $message = __('Cannot find order to ship');
            $this->logger->critical($message->render());
            throw new CouldNotShipException(new Phrase($message->render()));
        }

        $orderIncrementId = $shipmentExtensionAttributes->getOrderIncrementId();
        $carrierCode = $shipmentExtensionAttributes->getCarrierCode();
        if (!$carrierCode) {
            $carrierCode = 'ups';
        }
        $carrierTitle = $shipmentExtensionAttributes->getCarrierTitle();
        if (!$carrierTitle) {
            $carrierTitle = 'UPS';
        }
        $searchCriteria = $this->searchCriteriaBuilder
            ->addFilter('increment_id', $orderIncrementId, 'eq')->create();
        $orderList = $this->orderRepository->getList($searchCriteria);
        if (count($orderList) !== 1) {
            $message = __('Cannot find order %1', $orderIncrementId);
            $this->logger->critical($message->render());
            throw new CouldNotShipException(new Phrase($message->render()));
        }
        try {
            /** @var OrderInterface $order */
            $order = $orderList->getFirstItem();
            $tracks = $shipment_data->getTracks();

            $shipmentTracks = [];
            $numbers = [];
            if ($tracks) {
                foreach ($tracks as $track) {
                    $shipmentTrack = $this->shipmentTrackInterfaceFactory->create();
                    $shipmentTrack->setOrderId($order->getId());
                    $shipmentTrack->setCarrierCode($carrierCode);
                    $shipmentTrack->setTitle($carrierTitle);
                    $shipmentTrack->setTrackNumber($track->getTrackNumber());
                    $numbers[] = $track->getTrackNumber();
                    $shipmentTracks[] = $shipmentTrack;
                }
            }

            $this->createShipment($order, $shipmentTracks);

            $this->response->setStatus('success');
            $this->response->setMessage(__('Order %1 has been shipped:  %2', $order->getIncrementId(), implode(',', $numbers)));
            $this->response->setFlow('update_shipment');
            $this->response->setOrderIncrementId($order->getIncrementId());
            return $this->response;
        } catch (Exception $e) {
            $this->logger->critical($e->getMessage(), ['exception', $e]);
            throw $e;
        }
    }

    /**
     * Create shipment
     *
     * @param OrderInterface $order
     * @param ShipmentTrackInterface [] $shipmentTracks
     * @throws CouldNotShipException
     * @throws MailException
     * @throws LocalizedException
     * @see \Magento\Sales\Model\ShipOrder
     */
    protected function createShipment(OrderInterface $order, $shipmentTracks)
    {
        if (!$order->canShip()) {
            throw new CouldNotShipException(new Phrase('Order can not be shipped.'));
        }

        $objectManager = ObjectManager::getInstance();

        $convertOrder = $objectManager->create(Order::class);
        $shipment = $convertOrder->toShipment($order);

        foreach ($shipmentTracks as $track) {
            $shipment->addTrack($track);
        }

        $shippedItems = [];

        foreach ($order->getAllItems() as $orderItem) {
            if ($orderItem->getProductType() === Type::DEFAULT_TYPE) {
                $shippedItems[$orderItem->getProduct()->getSku()] = $orderItem->getSimpleQtyToShip();
            }

            if (!$orderItem->getQtyToShip() || $orderItem->getIsVirtual()) {
                continue;
            }

            $qtyShipped = $orderItem->getQtyToShip();
            $shipmentItem = $convertOrder->itemToShipmentItem($orderItem)->setQty($qtyShipped);
            $shipment->addItem($shipmentItem);
        }

        $this->addMultiStockInformationIfNeeded($shipment, $shippedItems);

        $connection = $this->resourceConnection->getConnection('sales');
        $connection->beginTransaction();

        try {
            $this->orderRegistrar->register($order, $shipment);
            $order->setState(
                $this->orderStateResolver->getStateForOrder($order, [OrderStateResolverInterface::IN_PROGRESS])
            );

            $order->setStatus($this->config->getStateDefaultStatus($order->getState()));
            $this->shipmentRepository->save($shipment);

            if ($shipmentTracks && !empty($shipmentTracks)) {
                foreach ($shipmentTracks as $track) {
                    $order->addStatusHistoryComment(__('Order has been shipped: %1 ', $track->getTrackNumber()), 'shipped');
                }
            } else {
                $order->addStatusHistoryComment(__('Order has been shipped: no tracking information'), 'shipped');
            }

            $order->addStatusHistoryComment(__('Order complete'), 'complete');
            $order->setStatus('complete');
            $order->setState('complete');
            $this->orderRepository->save($order);
            $connection->commit();
        } catch (Exception $e) {
            $this->logger->critical($e);
            $connection->rollBack();
            throw new CouldNotShipException(
                __('Could not save a shipment, see error log for details')
            );
        }

       // $this->shipmentNotifier->notify($shipment);
    }

    /**
     * @param ShipmentInterface $shipment
     * @param array $shippedItems
     * @throws CouldNotShipException
     * @throws Exception
     */
    protected function addMultiStockInformationIfNeeded(ShipmentInterface $shipment, array $shippedItems)
    {
        $shipmentItems = $shipment->getItems();

        if (!$shipmentItems || empty($shipmentItems)) {
            throw new CouldNotShipException(new Phrase('Empty shipment'));
        }

        if (version_compare($this->productMetadata->getVersion(), '2.3.0', '<')) {
            return;
        }

        $objectManager = ObjectManager::getInstance();

        $sourceSelection = $objectManager->create(GetDefaultSourceSelectionAlgorithmCodeInterface::class);

        $itemRequestFactory = $objectManager->create(ItemRequestInterfaceFactory::class);
        $inventoryRequestFactory = $objectManager->create(InventoryRequestInterfaceFactory::class);
        $sourceSelectionService = $objectManager->create(SourceSelectionServiceInterface::class);
        $storeManager = $objectManager->create(StoreManagerInterface::class);
        $websiteId = $storeManager->getStore($shipment->getStoreId())->getWebsiteId();
        $websiteCode = $storeManager->getWebsite($websiteId)->getCode();

        /** @var StockIdResolver $stockIdResolver */
        $stockIdResolver = $objectManager->create(StockIdResolver::class);
        /** @var StockInterface $stock */
        $stockId = $stockIdResolver->resolve(SalesChannelInterface::TYPE_WEBSITE, $websiteCode);
        $algorithmCode = $sourceSelection->execute();

        $items = [];
        foreach ($shippedItems as $sku => $qty) {
            $items[] = $itemRequestFactory->create([
                'sku' => $sku,
                'qty' => $qty
            ]);
        }

        $inventoryRequest = $inventoryRequestFactory->create([
            'stockId' => $stockId,
            'items' => $items
        ]);

        $sourceSelectionResult = $sourceSelectionService->execute(
            $inventoryRequest,
            $algorithmCode
        );

        // TODO: re-enable when fixed
        // if (!$sourceSelectionResult->isShippable()) {
        //      throw new CouldNotShipException(new Phrase('No quantity available for shipment'));
        // }

        foreach ($sourceSelectionResult->getSourceSelectionItems() as $product) {
            /** @var ShipmentExtensionInterface $extensionAttributes */
            $extensionAttributes = $shipment->getExtensionAttributes();
            $extensionAttributes->setSourceCode($product->getSourceCode());
            $shipment->setExtensionAttributes($extensionAttributes);
        }
    }
}
